import signal
import pytest
import json
import os

from typing import Generator

from tarantool.error import (  # type: ignore
    NetworkError,
)

from conftest import (
    MAX_LOGIN_ATTEMPTS,
    Instance,
    Cluster,
    Retriable,
)


@pytest.fixture
def instance_with_audit_file(unstarted_instance: Instance):
    unstarted_instance.audit = os.path.join(unstarted_instance.instance_dir, "audit.log")
    unstarted_instance.start()
    unstarted_instance.wait_online()
    yield unstarted_instance


class AuditFile:
    def __init__(self, path):
        self._f = open(path)

    def events(self) -> Generator[dict, None, None]:
        for line in self._f:
            yield json.loads(line)


def take_until_title(events, title: str) -> dict | None:
    for event in events:
        if event["title"] == title:
            return event
    return None


def test_startup(instance_with_audit_file: Instance):
    instance = instance_with_audit_file
    instance.start()
    instance.terminate()

    events = list(AuditFile(instance.audit_flag_value).events())
    assert len(events) > 0

    # Check identifiers
    i = 1
    event: dict | None
    for event in events:
        assert event["id"] == f"1.0.{i}"
        i += 1

    # These should be the first two events
    assert events[0]["title"] == "init_audit"
    assert events[0]["message"] == "audit log is ready"
    assert events[0]["severity"] == "low"
    assert events[1]["title"] == "local_startup"
    assert events[1]["message"] == "instance is starting"
    assert events[1]["severity"] == "low"

    event = take_until_title(iter(events), "join_instance")
    assert event is not None
    assert event["instance_name"] == "default_1_1"
    assert event["raft_id"] == "1"
    assert event["initiator"] == "admin"

    event = take_until_title(iter(events), "change_target_state")
    assert event is not None
    assert event["new_state"] == "Offline(0)"
    assert event["instance_name"] == "default_1_1"
    assert event["raft_id"] == "1"
    assert event["message"] == f"target state of instance `{event['instance_name']}` changed to {event['new_state']}"
    assert event["severity"] == "low"
    assert event["initiator"] == "admin"

    event = take_until_title(iter(events), "change_current_state")
    assert event is not None
    assert event["new_state"] == "Offline(0)"
    assert event["instance_name"] == "default_1_1"
    assert event["raft_id"] == "1"
    assert event["message"] == f"current state of instance `{event['instance_name']}` changed to {event['new_state']}"
    assert event["severity"] == "medium"
    assert event["initiator"] == "admin"

    create_db = take_until_title(events, "create_local_db")
    assert create_db is not None
    assert create_db["initiator"] == "admin"
    assert create_db["instance_name"] == "default_1_1"
    assert create_db["raft_id"] == "1"

    event = take_until_title(events, "connect_local_db")
    assert event is not None
    assert event["initiator"] == "admin"
    assert event["instance_name"] == "default_1_1"
    assert event["raft_id"] == "1"


def test_integrity_violation(instance_with_audit_file: Instance):
    instance = instance_with_audit_file
    # Instance was up for some time
    instance.start()
    instance.terminate()

    # Then data files were modified and instance restarted
    snap_name = f"{instance.instance_dir}/00000000000000000000.snap"
    with open(snap_name, "w") as snap:
        snap.write("abc")

    # Instance should fail to start
    instance.fail_to_start()

    # Instance should  detect integrity violation
    events = list(AuditFile(instance.audit_flag_value).events())
    event = take_until_title(iter(events), "integrity_violation")
    assert event is not None
    assert event["message"] == "integrity violation detected"
    assert event["severity"] == "high"
    assert event["error"] == "XlogError: Unexpected end of file, run with 'force_recovery = true'"


def test_recover_database(instance_with_audit_file: Instance):
    instance = instance_with_audit_file
    instance.start()
    instance.wait_online()

    instance.restart()
    instance.wait_online()
    instance.terminate()

    events = AuditFile(instance.audit_flag_value).events()

    # At first local db (data storage) is initialized
    create_db = take_until_title(events, "create_local_db")
    assert create_db is not None
    assert create_db["initiator"] == "admin"
    assert create_db["instance_name"] == "default_1_1"
    assert create_db["raft_id"] == "1"

    # On restart instance recovers it's local data
    event = take_until_title(iter(events), "recover_local_db")
    assert event is not None
    assert event["initiator"] == "admin"
    assert event["instance_name"] == "default_1_1"
    assert event["raft_id"] == "1"

    event = take_until_title(events, "connect_local_db")
    assert event is not None
    assert event["initiator"] == "admin"
    assert event["instance_name"] == "default_1_1"
    assert event["raft_id"] == "1"


def test_create_drop_table(instance_with_audit_file: Instance):
    instance = instance_with_audit_file
    instance.start()
    instance.sql(
        """
        create table "foo" ("val" int not null, primary key ("val"))
        distributed by ("val")
        """
    )
    instance.sql(
        """
        drop table "foo"
        """
    )
    instance.terminate()

    events = AuditFile(instance.audit_flag_value).events()

    create_table = take_until_title(events, "create_table")
    assert create_table is not None
    assert create_table["name"] == "foo"
    assert create_table["message"] == "created table `foo`"
    assert create_table["severity"] == "medium"
    assert create_table["initiator"] == "pico_service"

    drop_table = take_until_title(events, "drop_table")
    assert drop_table is not None
    assert drop_table["name"] == "foo"
    assert drop_table["message"] == "dropped table `foo`"
    assert drop_table["severity"] == "medium"
    assert drop_table["initiator"] == "pico_service"


def test_user(instance_with_audit_file: Instance):
    instance = instance_with_audit_file
    instance.start()
    instance.sql(
        """
        create user "ymir" with password 'T0psecret' using chap-sha1
        """
    )
    # TODO user cant change password without access to _pico_property
    # https://git.picodata.io/picodata/picodata/picodata/-/issues/449
    instance.sql(
        """
        alter user "ymir" password 'Topsecre1' using chap-sha1
        """,
        sudo=True,
    )
    instance.sql(
        """
        drop user "ymir"
        """
    )
    instance.terminate()

    events = AuditFile(instance.audit_flag_value).events()

    create_user = take_until_title(events, "create_user")
    assert create_user is not None
    assert create_user["user"] == "ymir"
    assert create_user["auth_type"] == "chap-sha1"
    assert create_user["message"] == f"created user `{create_user['user']}`"
    assert create_user["severity"] == "high"
    assert create_user["initiator"] == "pico_service"

    change_password = take_until_title(events, "change_password")
    assert change_password is not None
    assert change_password["user"] == "ymir"
    assert change_password["auth_type"] == "chap-sha1"
    assert change_password["message"] == f"password of user `{change_password['user']}` was changed"
    assert change_password["severity"] == "high"
    assert change_password["initiator"] == "admin"

    drop_user = take_until_title(events, "drop_user")
    assert drop_user is not None
    assert drop_user["user"] == "ymir"
    assert drop_user["message"] == f"dropped user `{drop_user['user']}`"
    assert drop_user["severity"] == "medium"
    assert drop_user["initiator"] == "pico_service"


def test_role(instance_with_audit_file: Instance):
    instance = instance_with_audit_file
    instance.start()

    setup = [
        """
        create user "bubba" with password 'T0psecret' using chap-sha1
        """,
        """
        grant create role to "bubba"
        """,
    ]
    for query in setup:
        instance.sql(query, sudo=True)

    with instance.connect(timeout=1, user="bubba", password="T0psecret") as c:
        c.sql(
            """
            create role "skibidi"
            """
        )
        c.sql(
            """
            create role "dummy"
            """
        )
        c.sql(
            """
            grant "dummy" to "skibidi"
            """
        )
        c.sql(
            """
            revoke "dummy" from "skibidi"
            """
        )
        c.sql(
            """
            drop role "skibidi"
            """
        )
    instance.terminate()

    events = AuditFile(instance.audit_flag_value).events()

    create_role = take_until_title(events, "create_role")
    assert create_role is not None
    assert create_role["role"] == "skibidi"
    assert create_role["message"] == f"created role `{create_role['role']}`"
    assert create_role["severity"] == "high"
    assert create_role["initiator"] == "bubba"

    grant_role = take_until_title(events, "grant_role")
    assert grant_role is not None
    assert grant_role["role"] == "dummy"
    assert grant_role["grantee"] == "skibidi"
    assert grant_role["grantee_type"] == "role"
    assert grant_role["message"] == f"granted role `{grant_role['role']}` to role `{grant_role['grantee']}`"
    assert grant_role["severity"] == "high"
    assert grant_role["initiator"] == "bubba"

    revoke_role = take_until_title(events, "revoke_role")
    assert revoke_role is not None
    assert revoke_role["role"] == "dummy"
    assert revoke_role["grantee"] == "skibidi"
    assert revoke_role["grantee_type"] == "role"
    assert revoke_role["message"] == f"revoked role `{grant_role['role']}` from role `{revoke_role['grantee']}`"
    assert revoke_role["severity"] == "high"
    assert revoke_role["initiator"] == "bubba"

    drop_role = take_until_title(events, "drop_role")
    assert drop_role is not None
    assert drop_role["role"] == "skibidi"
    assert drop_role["message"] == f"dropped role `{drop_role['role']}`"
    assert drop_role["severity"] == "medium"
    assert drop_role["initiator"] == "bubba"


def test_index(instance_with_audit_file: Instance):
    instance = instance_with_audit_file
    instance.start()
    instance.sql(
        """
        create table "foo" ("val" int not null, primary key ("val"))
        distributed by ("val")
        """
    )
    instance.sql(""" create index "foo_idx" on "foo" ("val") """)
    instance.sql(""" drop index "foo_idx" """)
    instance.terminate()

    events = AuditFile(instance.audit_flag_value).events()

    create_index = take_until_title(events, "create_index")
    assert create_index is not None
    assert create_index["name"] == "foo_idx"
    assert create_index["message"] == f"created index `{create_index['name']}`"
    assert create_index["severity"] == "medium"
    assert create_index["initiator"] == "pico_service"

    drop_index = take_until_title(events, "drop_index")
    assert drop_index is not None
    assert drop_index["name"] == "foo_idx"
    assert drop_index["message"] == f"dropped index `{drop_index['name']}`"
    assert drop_index["severity"] == "medium"
    assert drop_index["initiator"] == "pico_service"


def test_join_expel_instance(cluster: Cluster):
    cluster.deploy(instance_count=0)
    audit = os.path.join(cluster.data_dir, "audit.log")
    i1 = cluster.add_instance(audit=audit, replicaset_name="default_1")

    audit_i1 = AuditFile(i1.audit_flag_value)
    for _ in audit_i1.events():
        pass
    events = audit_i1.events()

    audit = os.path.join(cluster.data_dir, "audit.log")
    i2 = cluster.add_instance(name="i2", audit=audit, replicaset_name="default_1")

    join_instance = take_until_title(events, "join_instance")
    assert join_instance is not None
    assert join_instance["instance_name"] == "i2"
    assert join_instance["raft_id"] == str(i2.raft_id)
    assert join_instance["severity"] == "low"
    assert join_instance["initiator"] == "admin"

    # Joining of a new instance is treated as creation of a local database on it
    create_db = take_until_title(events, "create_local_db")
    assert create_db is not None
    assert create_db["instance_name"] == "i2"
    assert create_db["raft_id"] == str(i2.raft_id)
    assert create_db["severity"] == "low"
    assert create_db["initiator"] == "admin"

    cluster.expel(i2, force=True)
    cluster.wait_has_states(i2, "Expelled", "Expelled")

    expel_instance = take_until_title(events, "expel_instance")
    assert expel_instance is not None
    assert expel_instance["instance_name"] == "i2"
    assert expel_instance["raft_id"] == str(i2.raft_id)
    assert expel_instance["severity"] == "low"
    assert expel_instance["initiator"] == "admin"

    # Expelling an instance is treated as removal of a local database on it
    drop_db = take_until_title(events, "drop_local_db")
    assert drop_db is not None
    assert drop_db["instance_name"] == "i2"
    assert drop_db["raft_id"] == str(i2.raft_id)
    assert drop_db["severity"] == "low"
    assert drop_db["initiator"] == "admin"


def test_join_connect_instance(cluster: Cluster):
    cluster.deploy(instance_count=0)
    audit = os.path.join(cluster.data_dir, "audit.log")
    i1 = cluster.add_instance(audit=audit)

    audit = os.path.join(cluster.data_dir, "audit.log")
    i2 = cluster.add_instance(name="i2", audit=audit)
    i2.terminate()

    events = AuditFile(i2.audit_flag_value).events()

    create_db = take_until_title(events, "create_local_db")
    assert create_db is not None
    assert create_db["instance_name"] == "default_1_1"
    assert create_db["raft_id"] == str(i1.raft_id)
    assert create_db["severity"] == "low"
    assert create_db["initiator"] == "admin"

    create_db = take_until_title(events, "create_local_db")
    assert create_db is not None
    assert create_db["instance_name"] == "i2"
    assert create_db["raft_id"] == str(i2.raft_id)
    assert create_db["severity"] == "low"
    assert create_db["initiator"] == "admin"

    # Joining instance should emit 'connect_local_db' audit event.
    # connect_local_db is a local event as only joining instance emits it
    # in contrast to create_local_db
    connect_db = take_until_title(events, "connect_local_db")
    assert connect_db is not None
    assert connect_db["instance_name"] == "i2"
    assert connect_db["raft_id"] == str(i2.raft_id)
    assert connect_db["severity"] == "low"
    assert connect_db["initiator"] == "admin"


def test_auth(instance_with_audit_file: Instance):
    instance = instance_with_audit_file
    instance.start()

    instance.sql(
        """
        create user "ymir" with password 'T0psecret' using chap-sha1
        """,
        sudo=True,
    )
    instance.sql(
        """
        alter user "ymir" login
        """,
        sudo=True,
    )

    audit = AuditFile(instance.audit_flag_value)
    for _ in audit.events():
        pass
    events = audit.events()

    with instance.connect(4, user="ymir", password="T0psecret"):
        pass

    auth_ok = take_until_title(events, "auth_ok")
    assert auth_ok is not None
    assert auth_ok["user"] == "ymir"
    assert auth_ok["severity"] == "high"
    assert auth_ok["initiator"] == "ymir"
    assert auth_ok["verdict"] == "user is not blocked"

    for _ in range(MAX_LOGIN_ATTEMPTS):
        with pytest.raises(NetworkError, match="User not found or supplied credentials are invalid"):
            with instance.connect(4, user="ymir", password="wrong_pwd"):
                pass

        auth_fail = take_until_title(events, "auth_fail")
        assert auth_fail is not None
        assert auth_fail["message"] == "failed to authenticate user `ymir`"
        assert auth_fail["severity"] == "high"
        assert auth_fail["verdict"] == "user is not blocked"
        assert auth_fail["user"] == "ymir"
        assert auth_fail["initiator"] == "ymir"

    with pytest.raises(NetworkError, match="Maximum number of login attempts exceeded"):
        with instance.connect(4, user="ymir", password="Wr0ng_pwd"):
            pass

    auth_fail = take_until_title(events, "auth_fail")
    assert auth_fail is not None
    assert auth_fail["message"] == "failed to authenticate user `ymir`"
    assert auth_fail["severity"] == "high"
    assert auth_fail["verdict"] == ("Maximum number of login attempts exceeded; user blocked")
    assert auth_fail["user"] == "ymir"
    assert auth_fail["initiator"] == "ymir"


def test_access_denied(instance_with_audit_file: Instance):
    instance = instance_with_audit_file
    instance.start()

    instance.create_user(with_name="ymir", with_password="T0psecret", with_auth="chap-sha1")

    audit = AuditFile(instance.audit_flag_value)
    for _ in audit.events():
        pass

    events = audit.events()

    expected_error = "Create access to role 'R' is denied for user 'ymir'"
    expected_audit = "create access denied on role `R` for user `ymir`"

    with pytest.raises(
        Exception,
        match=expected_error,
    ):
        instance.sql('CREATE ROLE "R"', user="ymir", password="T0psecret")

    access_denied = take_until_title(events, "access_denied")
    assert access_denied is not None
    assert access_denied["message"] == expected_audit
    assert access_denied["severity"] == "medium"
    assert access_denied["privilege"] == "create"
    assert access_denied["object_type"] == "role"
    assert access_denied["object"] == "R"
    assert access_denied["initiator"] == "ymir"


def test_grant_revoke(instance_with_audit_file: Instance):
    instance = instance_with_audit_file
    instance.start()

    user = "ymir"
    password = "T0psecret"

    instance.create_user(with_name=user, with_password=password, with_auth="chap-sha1")

    instance.sql(f'GRANT CREATE ROLE TO "{user}"', sudo=True)

    audit = AuditFile(instance.audit_flag_value)
    for _ in audit.events():
        pass

    events = audit.events()

    # wildcard privilege to user
    instance.sql(f'GRANT CREATE TABLE TO "{user}"', sudo=True)

    grant_privilege = take_until_title(events, "grant_privilege")

    assert grant_privilege is not None
    assert grant_privilege["message"] == "granted privilege create on table `*` to user `ymir`"
    assert grant_privilege["severity"] == "high"
    assert grant_privilege["privilege"] == "create"
    assert grant_privilege["object"] == "*"
    assert grant_privilege["object_type"] == "table"
    assert grant_privilege["grantee"] == user
    assert grant_privilege["grantee_type"] == "user"
    assert grant_privilege["initiator"] == "admin"

    instance.sql(f'REVOKE CREATE TABLE FROM "{user}"', sudo=True)

    revoke_privilege = take_until_title(events, "revoke_privilege")

    assert revoke_privilege is not None
    assert revoke_privilege["message"] == f"revoked privilege create on table `*` from user `{user}`"
    assert revoke_privilege["severity"] == "high"
    assert revoke_privilege["privilege"] == "create"
    assert revoke_privilege["object"] == "*"
    assert revoke_privilege["object_type"] == "table"
    assert revoke_privilege["grantee"] == user
    assert revoke_privilege["grantee_type"] == "user"
    assert revoke_privilege["initiator"] == "admin"

    # specific privilege to user
    instance.sql(f'GRANT READ ON TABLE "_pico_tier" TO "{user}"', sudo=True)

    grant_privilege = take_until_title(events, "grant_privilege")

    assert grant_privilege is not None
    assert grant_privilege["message"] == f"granted privilege read on table `_pico_tier` to user `{user}`"
    assert grant_privilege["severity"] == "high"
    assert grant_privilege["privilege"] == "read"
    assert grant_privilege["object_type"] == "table"
    assert grant_privilege["grantee"] == user
    assert grant_privilege["grantee_type"] == "user"
    assert grant_privilege["initiator"] == "admin"
    assert grant_privilege["object"] == "_pico_tier"

    instance.sql(f'REVOKE READ ON TABLE "_pico_tier" FROM "{user}"', sudo=True)

    revoke_privilege = take_until_title(events, "revoke_privilege")

    assert revoke_privilege is not None
    assert revoke_privilege["message"] == f"revoked privilege read on table `_pico_tier` from user `{user}`"
    assert revoke_privilege["severity"] == "high"
    assert revoke_privilege["privilege"] == "read"
    assert revoke_privilege["object_type"] == "table"
    assert revoke_privilege["grantee"] == user
    assert revoke_privilege["grantee_type"] == "user"
    assert revoke_privilege["initiator"] == "admin"
    assert revoke_privilege["object"] == "_pico_tier"

    instance.sql('CREATE ROLE "R"', user=user, password=password)

    # wildcard privilege to role
    instance.sql('GRANT CREATE TABLE TO "R"', sudo=True)

    grant_privilege = take_until_title(events, "grant_privilege")

    assert grant_privilege is not None
    assert grant_privilege["message"] == "granted privilege create on table `*` to role `R`"
    assert grant_privilege["severity"] == "high"
    assert grant_privilege["privilege"] == "create"
    assert grant_privilege["object"] == "*"
    assert grant_privilege["object_type"] == "table"
    assert grant_privilege["grantee"] == "R"
    assert grant_privilege["grantee_type"] == "role"
    assert grant_privilege["initiator"] == "admin"

    instance.sql('REVOKE CREATE TABLE FROM "R"', sudo=True)

    revoke_privilege = take_until_title(events, "revoke_privilege")

    assert revoke_privilege is not None
    assert revoke_privilege["message"] == "revoked privilege create on table `*` from role `R`"
    assert revoke_privilege["severity"] == "high"
    assert revoke_privilege["privilege"] == "create"
    assert grant_privilege["object"] == "*"
    assert revoke_privilege["object_type"] == "table"
    assert revoke_privilege["grantee"] == "R"
    assert revoke_privilege["grantee_type"] == "role"
    assert revoke_privilege["initiator"] == "admin"

    # specific privilege to role
    instance.sql('GRANT READ ON TABLE "_pico_user" TO "R"', sudo=True)

    grant_privilege = take_until_title(events, "grant_privilege")

    assert grant_privilege is not None
    assert grant_privilege["message"] == "granted privilege read on table `_pico_user` to role `R`"
    assert grant_privilege["severity"] == "high"
    assert grant_privilege["privilege"] == "read"
    assert grant_privilege["object_type"] == "table"
    assert grant_privilege["grantee"] == "R"
    assert grant_privilege["grantee_type"] == "role"
    assert grant_privilege["initiator"] == "admin"
    assert grant_privilege["object"] == "_pico_user"

    instance.sql('REVOKE READ ON TABLE "_pico_user" FROM "R"', sudo=True)

    revoke_privilege = take_until_title(events, "revoke_privilege")

    assert revoke_privilege is not None
    assert revoke_privilege["message"] == "revoked privilege read on table `_pico_user` from role `R`"
    assert revoke_privilege["severity"] == "high"
    assert revoke_privilege["privilege"] == "read"
    assert revoke_privilege["object_type"] == "table"
    assert revoke_privilege["grantee"] == "R"
    assert revoke_privilege["grantee_type"] == "role"
    assert revoke_privilege["initiator"] == "admin"
    assert revoke_privilege["object"] == "_pico_user"

    # role to user
    instance.sql('GRANT "R" TO "ymir"', user=user, password=password)

    grant_role = take_until_title(events, "grant_role")
    assert grant_role is not None
    assert grant_role["message"] == "granted role `R` to user `ymir`"
    assert grant_role["severity"] == "high"
    assert grant_role["role"] == "R"
    assert grant_role["grantee"] == "ymir"
    assert grant_role["grantee_type"] == "user"
    assert grant_role["initiator"] == "ymir"

    instance.sql(f'REVOKE "R" FROM "{user}"', user=user, password=password)

    revoke_role = take_until_title(events, "revoke_role")

    assert revoke_role is not None
    assert revoke_role["message"] == f"revoked role `R` from user `{user}`"
    assert revoke_role["severity"] == "high"
    assert revoke_role["role"] == "R"
    assert revoke_role["grantee"] == user
    assert revoke_role["grantee_type"] == "user"
    assert revoke_role["initiator"] == user

    # one role to another role
    instance.sql('CREATE ROLE "R2"', user=user, password=password)
    instance.sql('GRANT "R" TO "R2"', user=user, password=password)

    grant_role = take_until_title(events, "grant_role")
    assert grant_role is not None
    assert grant_role["message"] == "granted role `R` to role `R2`"
    assert grant_role["severity"] == "high"
    assert grant_role["role"] == "R"
    assert grant_role["grantee"] == "R2"
    assert grant_role["grantee_type"] == "role"
    assert grant_role["initiator"] == "ymir"

    instance.sql('REVOKE "R" FROM "R2"', user=user, password=password)

    revoke_role = take_until_title(events, "revoke_role")

    assert revoke_role is not None
    assert revoke_role["message"] == "revoked role `R` from role `R2`"
    assert revoke_role["severity"] == "high"
    assert revoke_role["role"] == "R"
    assert revoke_role["grantee"] == "R2"
    assert revoke_role["grantee_type"] == "role"
    assert revoke_role["initiator"] == user


def check_rotate(audit: AuditFile):
    rotate = take_until_title(audit.events(), "audit_rotate")
    assert rotate is not None
    assert rotate["message"] == "log file has been reopened"
    assert rotate["severity"] == "low"


def test_rotation(instance_with_audit_file: Instance):
    instance = instance_with_audit_file
    instance.start()

    user = "ymir"
    password = "T0psecret"

    audit = AuditFile(instance.audit_flag_value)
    for _ in audit.events():
        pass

    instance.create_user(with_name=user, with_password=password)

    instance.sql(f'GRANT CREATE ROLE TO "{user}"', sudo=True)

    assert instance.process is not None
    instance.process.send_signal(sig=signal.SIGHUP)

    Retriable(timeout=5, rps=1).call(check_rotate, audit)


def test_rename_user(instance_with_audit_file: Instance):
    instance = instance_with_audit_file
    instance.start()

    user = "ymir"
    password = "T0psecret"

    instance.create_user(with_name=user, with_password=password)

    instance.sql('ALTER USER "ymir" RENAME TO "TOM"', sudo=True)

    instance.terminate()

    events = AuditFile(instance.audit_flag_value).events()

    rename_user = take_until_title(events, "rename_user")
    assert rename_user is not None
    assert rename_user["message"] == "name of user `ymir` was changed to `TOM`"
    assert rename_user["severity"] == "high"
    assert rename_user["old_name"] == "ymir"
    assert rename_user["new_name"] == "TOM"
